from typing import Optional, Any

import numpy as np
import pandas as pd
from skbase.base import BaseEstimator
from sklearn import clone
from sklearn.base import ClassifierMixin
from sklearn.linear_model import LogisticRegression




class SkLearnWrapper(BaseEstimator, ClassifierMixin):

    def __init__(self,
                  model: Any,
                 ):

        self.model = model
        self.classes_ = []
        super().__init__()

    def fit(self, X: pd.DataFrame, y: pd.Series):
        self.classes_ = np.sort(np.unique(y))
        self.model.fit(X, y)

    def predict_proba(self, X: pd.DataFrame) -> np.ndarray:
        return self.model.predict_proba(X)


class OrdinalClassifier(BaseEstimator, ClassifierMixin):

    def __init__(self,
                 estimator: Optional = None,
                 ):
        self.estimator = estimator or LogisticRegression()
        self.clfs = {}
        self.classes_ = []
        self.coef_ = []
        super().__init__()

    def fit(self, X, y):
        self.classes_ = np.sort(np.unique(y))
        if self.classes_.shape[0] <= 2:
            raise ValueError("OrdinalClassifier needs at least 3 classes")

        for i in range(self.classes_.shape[0] - 1):
            binary_y = (y > self.classes_[i]).astype(np.uint8)
            clf = clone(self.estimator)
            clf.fit(X, binary_y)
            self.clfs[i] = clf
            try:
                self.coef_.append(clf.coef_)
            except AttributeError:
                pass

        return self

    def predict_proba(self, X):
        clfs_probs = {k: self.clfs[k].predict_proba(X) for k in self.clfs}
        predicted = []
        for i, y in enumerate(self.classes_):
            if i == 0:

                predicted.append(1 - clfs_probs[i][:, 1])
            elif i in clfs_probs:

                predicted.append(clfs_probs[i - 1][:, 1] - clfs_probs[i][:, 1])
            else:

                predicted.append(clfs_probs[i - 1][:, 1])
        return np.vstack(predicted).T

    def predict(self, X):
        return np.argmax(self.predict_proba(X), axis=1)


from player_performance_ratings.tuner.utils import ParameterSearchRange
from sklearn.linear_model import LogisticRegression
from sklearn.preprocessing import StandardScaler
from venn_abers import VennAbersCalibrator

from player_performance_ratings.examples.utils import load_nba_game_player_data, load_nba_game_matchup_data

from player_performance_ratings import ColumnNames, LogLossScorer

from player_performance_ratings.consts import PredictColumnNames
from player_performance_ratings.predictor.estimators.classifier import SkLearnGameTeamPredictor
from player_performance_ratings.predictor.estimators.sklearn_models import SkLearnWrapper
from player_performance_ratings.ratings.enums import RatingColumnNames
from player_performance_ratings.ratings.match_rating import TeamRatingGenerator
from player_performance_ratings.ratings.opponent_adjusted_rating.start_rating_generator import StartRatingGenerator

from player_performance_ratings.transformations.pre_transformers import SkLearnTransformerWrapper, MinMaxTransformer
from player_performance_ratings.ratings.opponent_adjusted_rating.rating_generator import OpponentAdjustedRatingGenerator
from player_performance_ratings.tuner import MatchPredictorTuner
from player_performance_ratings.tuner.match_predictor_factory import MatchPredictorFactory


column_names = ColumnNames(
    team_id='team_id',
    match_id='game_id',
    start_date="start_date",
    player_id="player_id",
    performance="plus_minus_per_minute",
    participation_weight="participation_weight",

)
gm = load_nba_game_matchup_data()
df = load_nba_game_player_data()

df = df[df['game_id'].isin(gm['game_id'].unique().tolist())]


print(len(df['game_id'].unique()))
df[PredictColumnNames.TARGET] = df['won']
df = df.sort_values(by=[column_names.start_date, column_names.match_id, column_names.team_id, column_names.player_id])
df = df[df['game_minutes']>46]
df['plus_minus_per_minute'] = df['plus_minus'] / df['game_minutes']
df['participation_weight'] = df['minutes'] / df['game_minutes']

df = (
    df.assign(team_count=df.groupby('game_id')['team_id'].transform('nunique'))
    .loc[lambda x: x.team_count == 2]
    .drop(columns=['team_count'])
)


features = ["plus_minus_per_minute"]
standard_scaler = SkLearnTransformerWrapper(transformer=StandardScaler(), features=features)
pre_transformers = [
    standard_scaler,
    MinMaxTransformer(features=features),
]


start_rating_search_range = [
    ParameterSearchRange(
        name='league_quantile',
        type='uniform',
        low=0.04,
        high=.4,
    ),
    ParameterSearchRange(
        name='min_count_for_percentiles',
        type='int',
        low=50,
        high=400,
    ),
]

estimator = SkLearnWrapper(VennAbersCalibrator(estimator=LogisticRegression(), inductive=True, cal_size=0.2, random_state=101))

predictor = SkLearnGameTeamPredictor(model=estimator,features=[RatingColumnNames.RATING_DIFFERENCE], game_id_colum='game_id',
                                     team_id_column='team_id', weight_column='participation_weight')



rating_generator = OpponentAdjustedRatingGenerator(
    team_rating_generator=TeamRatingGenerator(
        start_rating_generator=StartRatingGenerator(
            team_weight=0,
        )
    )
)



match_predictor_factory = MatchPredictorFactory(
    column_names=column_names,
    rating_generators=rating_generator,
    pre_transformers=pre_transformers,
    predictor=predictor,
    train_split_date="2022-05-01"
)



tuner = MatchPredictorTuner(
    match_predictor_factory=match_predictor_factory,
    fit_best=True,
    scorer=LogLossScorer(pred_column=predictor.pred_column),
    rating_generator_tuners=rating_generator_tuner,

)
best_match_predictor = tuner.tune(df=df)
tuner.tune(df=df)



import logging
from dataclasses import dataclass
from typing import Optional,  List

import numpy as np
import pandas as pd
from lightgbm import LGBMRegressor

from player_performance_ratings.transformation.base_transformer import BaseTransformer


@dataclass
class ColumnWeight:
    name: str
    weight: float
    lower_is_better: bool = False


class ColumnsWeighter(BaseTransformer):

    def __init__(self,
                 weighted_column_name: str,
                 column_weights: list[ColumnWeight]
                 ):
        self.weighted_column_name = weighted_column_name
        self.column_weights = column_weights

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        df[self.weighted_column_name] = 0

        df['sum_cols_weights'] = 0
        for column_weight in self.column_weights:
            df[f'weight__{column_weight.name}'] = column_weight.weight
            df.loc[df[column_weight.name].isna(), f'weight__{column_weight.name}'] = 0
            df.loc[df[column_weight.name].isna(), column_weight.name] = 0
            df['sum_cols_weights'] = df['sum_cols_weights'] + df[f'weight__{column_weight.name}']

        drop_cols = ['sum_cols_weights']
        for column_weight in self.column_weights:
            df[f'weight__{column_weight.name}'] / df['sum_cols_weights']
            drop_cols.append(f'weight__{column_weight.name}')

        for column_weight in self.column_weights:

            if column_weight.lower_is_better:
                df[self.weighted_column_name] += df[f'weight__{column_weight.name}'] * (1 - df[column_weight.name])
            else:
                df[self.weighted_column_name] += df[f'weight__{column_weight.name}'] * df[column_weight.name]
        df = df.drop(columns=drop_cols)
        return df

    @property
    def features_created(self) -> list[str]:
        return [self.weighted_column_name]


class SklearnEstimatorImputer(BaseTransformer):

    def __init__(self, features: list[str], target_name: str, estimator: Optional[LGBMRegressor] = None, ):
        self.estimator = estimator or LGBMRegressor()
        self.features = features
        self.target_name = target_name

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        self.estimator.fit(df[self.features], df[self.target_name])
        df = df.assign(**{
            f'imputed_col_{self.target_name}': self.estimator.predict(df[self.features])
        })
        df[self.target_name] = df[self.target_name].fillna(df[f'imputed_col_{self.target_name}'])
        return df.drop(columns=[f'imputed_col_{self.target_name}'])

    @property
    def features_created(self) -> list[str]:
        return [self.target_name]


class SkLearnTransformerWrapper(BaseTransformer):

    def __init__(self, transformer, features: list[str]):
        self.transformer = transformer
        self.features = features

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        df[self.features] = self.transformer.fit_transform(df[self.features])
        return df

    @property
    def features_created(self) -> list[str]:
        return self.features


class MinMaxTransformer(BaseTransformer):

    def __init__(self,
                 features: list[str],
                 quantile: float = 0.99,
                 allowed_mean_diff: Optional[float] = 0.01,
                 prefix: str = ""
                 ):
        self.features = features
        self.quantile = quantile
        self.allowed_mean_diff = allowed_mean_diff
        self.prefix = prefix
        if self.quantile < 0 or self.quantile > 1:
            raise ValueError("quantile must be between 0 and 1")

        self._feature_names_created = []

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        for feature in self.features:
            max_value = df[feature].quantile(self.quantile)
            min_value = df[feature].quantile(1 - self.quantile)

            df[self.prefix + feature] = (df[feature] - min_value) / (max_value - min_value)
            df[self.prefix + feature].clip(0, 1, inplace=True)
            self._feature_names_created.append(self.prefix + feature)

            if self.allowed_mean_diff:
                reps = 0
                mean_value = df[self.prefix + feature].mean()
                while abs(0.5 - mean_value) > self.allowed_mean_diff:

                    if mean_value > 0.5:
                        df[self.prefix + feature] = df[self.prefix + feature] * (1- self.allowed_mean_diff)
                    else:
                        df[self.prefix + feature] = df[self.prefix + feature] * (1+ self.allowed_mean_diff)

                    df[self.prefix + feature].clip(0, 1, inplace=True)
                    mean_value = df[self.prefix + feature].mean()

                    reps +=1
                    if reps > 100:
                        logging.warning(f"MinMaxTransformer: {feature} mean value is {mean_value} after {reps} repetitions")
                        continue

        return df

    @property
    def features_created(self) -> list[str]:
        return self._feature_names_created


class DiminishingValueTransformer(BaseTransformer):

    def __init__(self,
                 features: List[str],
                 cutoff_value: Optional[float] = None,
                 quantile_cutoff: float = 0.95,
                 excessive_multiplier: float = 0.8,

                 ):
        self.features = features
        self.cutoff_value = cutoff_value
        self.excessive_multiplier = excessive_multiplier
        self.trained_count: int = 0
        self.quantile_cutoff = quantile_cutoff

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:

        for feature_name in self.features:
            if self.cutoff_value is None:
                cutoff_value = df[feature_name].quantile(self.quantile_cutoff)
            else:
                cutoff_value = self.cutoff_value

            df = df.assign(feature_name=lambda x: np.where(
                x[feature_name] >= cutoff_value,
                (x[feature_name] - cutoff_value).clip(lower=0) * self.excessive_multiplier + cutoff_value,
                np.where(
                    x[feature_name] < -cutoff_value,
                    (-x[feature_name] - cutoff_value).clip(upper=0) * self.excessive_multiplier - cutoff_value,
                    x[feature_name]
                )
            ))

        df[self.features] = df[self.features].fillna(df[self.features])

        return df

    @property
    def features_created(self) -> list[str]:
        return self.features


from abc import ABC, abstractmethod
from typing import  Optional

import numpy as np
import pandas as pd
from sklearn.metrics import log_loss

from player_performance_ratings.consts import PredictColumnNames
from player_performance_ratings.ratings.enums import RatingColumnNames


class BaseScorer(ABC):

    def __init__(self, target: str, pred_column: str):
        self.target = target
        self.pred_column = pred_column

    @abstractmethod
    def score(self, df: pd.DataFrame, classes_: Optional[list[str]] = None) -> float:
        pass


class LogLossScorer(BaseScorer):

    def __init__(self, pred_column: str, target: Optional[str] = PredictColumnNames.TARGET, weight_cross_league: float = 1):
        self.pred_column_name = pred_column
        self.weight_cross_league = weight_cross_league
        super().__init__(target=target, pred_column=pred_column)

    def score(self, df: pd.DataFrame, classes_: Optional[list[str]] = None) -> float:
        if self.weight_cross_league == 1:
            return log_loss(df[self.target], df[self.pred_column_name])

        else:
            cross_league_rows = df[df[RatingColumnNames.PLAYER_LEAGUE] != RatingColumnNames.OPPONENT_LEAGUE]
            same_league_rows = df[df[RatingColumnNames.PLAYER_LEAGUE] == df[RatingColumnNames.OPPONENT_LEAGUE]]
            cross_league_logloss = log_loss(cross_league_rows[self.target], cross_league_rows[self.pred_column_name])
            same_league_logloss = log_loss(same_league_rows[self.target], same_league_rows[self.pred_column_name])

            weight_cross_league = len(cross_league_rows) * self.weight_cross_league / (
                    len(same_league_rows) + len(cross_league_rows) * self.weight_cross_league)

            return weight_cross_league * cross_league_logloss + (1 - weight_cross_league) * same_league_logloss

class OrdinalLossScorer(BaseScorer):

    def __init__(self,
                 pred_column: str,
                 target: Optional[str] = PredictColumnNames.TARGET,
                 granularity: Optional[list[str]] = None,
                 ):

        self.pred_column_name = pred_column
        self.granularity = granularity
        super().__init__(target=target, pred_column=pred_column)

    def score(self, df: pd.DataFrame, classes_: Optional[list[str]] = None) -> float:


        if classes_ is None:
            raise ValueError("classes_ must be passed to OrdinalLossScorer")

        probs = df[self.pred_column_name]
        last_column_name = 'prob_under_0.5'
        df[last_column_name] = probs.apply(lambda x: x[0])

        class_index = 0

        sum_lr = 0

        for class_ in classes_[1:]:
            class_index += 1
            p_c = 'prob_under_' + str(class_ + 0.5)
            df[p_c] = probs.apply(lambda x: x[class_index]) + df[last_column_name]

            count_exact = len(df[df['__target'] == class_])
            weight_class = count_exact / len(df)

            if self.granularity:
                grouped = df.groupby(self.granularity  +['__target'])[p_c].mean().reset_index()
            else:
                grouped = df

            grouped['min'] = 0.0001
            grouped['max'] = 0.9999
            grouped[p_c] = np.minimum(grouped['max'], grouped[p_c])
            grouped[p_c] = np.maximum(grouped['min'], grouped[p_c])
            grouped['log_loss'] = 0

            grouped.loc[grouped['__target'] <= class_, 'log_loss'] = np.log(grouped[p_c])
            grouped.loc[grouped['__target'] > class_, 'log_loss'] = np.log(1 - grouped[p_c])
            sum_lr -= grouped['log_loss'].mean() * weight_class

            last_column_name = p_c

        return sum_lr




from typing import Optional, List, Union

import pendulum

from player_performance_ratings import ColumnNames
from player_performance_ratings.predictor import MatchPredictor
from player_performance_ratings.predictor.estimators.base_estimator import BaseMLWrapper
from player_performance_ratings.ratings.opponent_adjusted_rating.rating_generator import RatingGenerator
from player_performance_ratings.transformation.base_transformer import BaseTransformer


class MatchPredictorFactory():

    def __init__(self,
                 column_names: Union[ColumnNames, list[ColumnNames]],
                 rating_generators: Union[RatingGenerator, list[RatingGenerator]],
                 pre_transformers: Optional[List[BaseTransformer]] = None,
                 post_transformers: Optional[List[BaseTransformer]] = None,
                 predictor: BaseMLWrapper = None,
                 train_split_date: Optional[pendulum.datetime] = None
                 ):

        self.rating_generators = rating_generators
        if isinstance(self.rating_generators, RatingGenerator):
            self.rating_generators = [self.rating_generators]

        if len(self.rating_generators) > 1 and predictor is None:
            raise ValueError(
                "If multiple rating generators are used, a predictor must be specified."
                " Otherwise it is not clear which features from which rating-model is used ")

        self.pre_transformers = pre_transformers or []
        self.post_transformers = post_transformers or []
        self.column_names = column_names
        self.predictor = predictor
        self.train_split_date = train_split_date

    def create(self,
               pre_rating_transformers: Optional[List[BaseTransformer]] = None,
               rating_generators: Optional[list[RatingGenerator]] = None,
               post_rating_transformers: Optional[List[BaseTransformer]] = None) -> MatchPredictor:

        rating_generators = rating_generators if rating_generators is not None else self.rating_generators
        pre_rating_transformers = pre_rating_transformers if pre_rating_transformers is not None else self.pre_transformers
        post_rating_transformers = post_rating_transformers if post_rating_transformers is not None else self.post_transformers

        return MatchPredictor(column_names=self.column_names, rating_generators=rating_generators,
                              pre_rating_transformers=pre_rating_transformers,
                              post_rating_transformers=post_rating_transformers,
                              predictor=self.predictor,
                              train_split_date=self.train_split_date)



from .performance_predictor import RatingDifferencePerformancePredictor, RatingMeanPerformancePredictor
from .rating_generator import OpponentAdjustedRatingGenerator
from .start_rating_generator import StartRatingGenerator
from .team_rating_generator import TeamRatingGenerator


import math

from typing import List, Dict

from player_performance_ratings.data_structures import Match


class LeagueIdentifier():

    def __init__(self,
                 matches_back: int = 25
                 ):
        self.matches_back = matches_back
        self.entity_to_match_leagues: Dict[str, List[str]] = {}
        self.entity_to_match_league_counts: Dict[str, Dict[str, int]] = {}
        self.entity_id_to_most_league_count: Dict[str, int] = {}
        self.entity_id_to_most_league_name: Dict[str, str] = {}


    def identify(self, player_id: str, league_match: str) -> str:
        if player_id not in self.entity_to_match_leagues:
            self.entity_to_match_leagues[player_id] = []
            self.entity_to_match_league_counts[player_id] = {}
            self.entity_id_to_most_league_count[player_id] = 0
            self.entity_id_to_most_league_name[player_id] = ""

        self.entity_to_match_leagues[player_id].append(league_match)
        if league_match not in self.entity_to_match_league_counts[player_id]:
            self.entity_to_match_league_counts[player_id][league_match] = 0

        if len(self.entity_to_match_leagues[player_id]) > self.matches_back:
            league_drop_out = self.entity_to_match_leagues[player_id][0]
            self.entity_to_match_league_counts[player_id][league_drop_out] -= 1
            self.entity_to_match_leagues[player_id] = self.entity_to_match_leagues[player_id][1:]
            if league_drop_out == self.entity_id_to_most_league_name[player_id]:
                self.entity_id_to_most_league_count[player_id] -= 1

        self.entity_to_match_league_counts[player_id][league_match] += 1
        if self.entity_to_match_league_counts[player_id][league_match] > self.entity_id_to_most_league_count[player_id]:
            self.entity_id_to_most_league_name[player_id] = league_match

        if self.entity_id_to_most_league_name[player_id] == league_match:
            self.entity_id_to_most_league_count[player_id] += 1

        return self.entity_id_to_most_league_name[player_id]



    def _generate_teams_to_leagues(self, team_ids: List[str], team_league_counts: Dict[str, Dict[str, int]]) -> Dict[
        str, str]:
        team_leagues: Dict[str, str] = {}
        for team_id in team_ids:
            league = self._identify_primary_league_for_team(team_league_counts[team_id])
            team_leagues[team_id] = league

        return team_leagues

    def _identify_primary_league_for_team(self, league_counts: Dict[str, int]) -> str:
        max_league: str = ""
        max_count = -math.inf
        for league, count in league_counts.items():
            if count > max_count:
                max_count = count
                max_league = league

        return max_league

    def _get_opponent_league(self, team_id: str, team_leagues: Dict[
        str, str]) -> str:

        for team_id2, league in team_leagues.items():
            if team_id2 != team_id:
                return team_leagues[team_id2]

        raise KeyError

    def _generate_team_league_counts(self, match: Match) -> Dict[str, Dict[str, int]]:

        team_league_counts: Dict[str, Dict[str, int]] = {}
        for team in match.teams:
            for player in team.players:
                if team.id not in team_league_counts:
                    team_league_counts[team.id] = {}

                if player.league not in team_league_counts[team.id]:
                    team_league_counts[team.id][player.league] = 0

                team_league_counts[team.id][player.league] += 1

        return team_league_counts

    def get_primary_league(self, match: Match) -> str:
        region_counts: dict[str, int] = {}
        max_count: int = 0
        primary_league: str = ""
        for team in match.teams:
            for player in team.players:
                region = player.league

                if region not in region_counts:
                    region_counts[region] = 0
                region_counts[region] += 1

                if region_counts[region] > max_count:
                    max_count = region_counts[region]
                    primary_league = region

        return primary_league



import numpy as np
from sklearn.preprocessing import StandardScaler

from player_performance_ratings.examples.utils import load_nba_game_matchup_data

from player_performance_ratings.consts import PredictColumnNames
from player_performance_ratings.data_structures import ColumnNames
from player_performance_ratings.predictor.match_predictor import MatchPredictor
from player_performance_ratings import RatingGenerator, SkLearnTransformerWrapper, \
    MinMaxTransformer, TeamRatingTuner, StartRatingTuner, MatchPredictorTuner, ParameterSearchRange
from player_performance_ratings.predictor.estimators.classifier import SkLearnGameTeamPredictor
from player_performance_ratings.predictor.estimators.ordinal_classifier import OrdinalClassifier
from player_performance_ratings.ratings.enums import RatingColumnNames
from player_performance_ratings.ratings.match_rating import TeamRatingGenerator
from player_performance_ratings.ratings.opponent_adjusted_rating.performance_predictor import RatingMeanPerformancePredictor
from player_performance_ratings.ratings.opponent_adjusted_rating.start_rating_generator import StartRatingGenerator
from player_performance_ratings.scorer.score import OrdinalLossScorer

column_names = ColumnNames(
    team_id='lineup_id',
    match_id='matchup_game_id',
    start_date="start_date",
    player_id="player_id",
    performance="total_points_lineup_matchup_per_minute",
    participation_weight="participation_weight",
    rating_update_id="game_id"

)

df = load_nba_game_matchup_data()
df['total_score'] = df['score'] + df['score_opponent']

df = df[df['game_minutes'] > 46]

df.loc[df['total_score'] > 248, 'total_score'] = 248
df.loc[df['total_score'] < 207, 'total_score'] = 207

df[PredictColumnNames.TARGET] = df['total_score']

min_lineup = np.minimum(df['lineup_id'], df['lineup_id_opponent'])
max_lineup = np.maximum(df['lineup_id'], df['lineup_id_opponent'])

# Combine the min and max values into a tuple
df['sorted_lineup'] = list(zip(min_lineup, max_lineup))

df[column_names.match_id] = df['game_id'].astype(str) + '_' + df['sorted_lineup'].astype(str)
df = df.drop(columns=['sorted_lineup'])

df = (
    df.assign(team_count=df.groupby(column_names.match_id)[column_names.team_id].transform('nunique'))
    .loc[lambda x: x.team_count == 2]
)
df = df.sort_values(by=[column_names.start_date, column_names.match_id, column_names.team_id, column_names.player_id])
df['total_points_lineup_matchup'] = df['points'] + df['points_opponent']
df['total_points_lineup_matchup_per_minute'] = df['total_points_lineup_matchup'] / df['minutes_lineup_matchup']
df.loc[df['minutes_lineup_matchup'] == 0, 'total_points_lineup_matchup_per_minute'] = 0

mean_participation_weight = df['participation_weight'].mean()
features = ['total_points_lineup_matchup_per_minute']

standard_scaler = SkLearnTransformerWrapper(transformer=StandardScaler(), features=features)
pre_transformer_search_ranges = [
    (standard_scaler, []),
    (MinMaxTransformer(features=features), [])
]

performance_predictor = RatingMeanPerformancePredictor(
)

rating_generator = RatingGenerator(
    store_game_ratings=True,
    column_names=column_names,
    team_rating_generator=TeamRatingGenerator(
        performance_predictor=performance_predictor,
        start_rating_generator=StartRatingGenerator(
            team_weight=0,
        )
    )
)

pre_transformers = [
    standard_scaler,
    MinMaxTransformer(features=features, quantile=0.99),
]

start_rating_search_range = [
    ParameterSearchRange(
        name='league_quantile',
        type='uniform',
        low=0.04,
        high=.5,
    ),
    ParameterSearchRange(
        name='min_count_for_percentiles',
        type='int',
        low=50,
        high=400,
    ),
]

predictor = SkLearnGameTeamPredictor(features=[RatingColumnNames.RATING_MEAN],
                                     weight_column='participation_weight',
                                     model=OrdinalClassifier(),
                                     team_id_column='team_id', game_id_colum=column_names.rating_update_id,
                                     target='total_score', multiclassifier=True)

scorer = OrdinalLossScorer(
    pred_column=predictor.pred_column,
    granularity=['game_id', 'team_id']
)

match_predictor = MatchPredictor(column_names=column_names, rating_generator=rating_generator, predictor=predictor,
                                 pre_rating_transformers=pre_transformers, train_split_date="2022-05-01")

team_rating_tuner = TeamRatingTuner(match_predictor=match_predictor,
                                    n_trials=35,
                                    scorer=scorer,
                                    )

start_rating_tuner = StartRatingTuner(column_names=column_names,
                                      match_predictor=match_predictor,
                                      search_ranges=start_rating_search_range,
                                      n_trials=9,
                                      scorer=scorer,
                                      )

tuner = MatchPredictorTuner(
    team_rating_tuner=team_rating_tuner,
    start_rating_tuner=start_rating_tuner,
    fit_best=True,

)
best_match_predictor = tuner.tune(df=df)
tuner.tune(df=df)



import logging
import warnings
import pandas as pd
from pandas.errors import SettingWithCopyWarning
from venn_abers import VennAbersCalibrator

warnings.simplefilter(action="ignore", category=SettingWithCopyWarning)

from typing import Optional

from sklearn.linear_model import LogisticRegression

from player_performance_ratings.consts import PredictColumnNames

from player_performance_ratings.data_structures import ColumnNames
from player_performance_ratings.predictor.estimators.base_estimator import BaseMLWrapper


class SkLearnGameTeamPredictor(BaseMLWrapper):

    def __init__(self,
                 game_id_colum: str,
                 team_id_column: str,
                 features: list[str],
                 weight_column: Optional[str] = None,
                 target: Optional[str] = PredictColumnNames.TARGET,
                 model: Optional = None,
                 multiclassifier: bool = False,
                 pred_column: Optional[str] = "prob",
                 ):
        """
        Wrapper for sklearn models that predicts game results.

        The GameTeam Predictor is intended to transform predictions from a lower granularity into a GameTeam level.
        So if input data is at game-player, data is converted to game_team before being trained
        Similar concept if it is at a granularity below game level.

        The weight_column makes it possible to weight certain rows more than others.
        For instance, if data is on game-player level and the participation rate of players is not equal -->
         setting participation_weight equal to weight_column will the feature values of the players with high participation_rates have higher weight before aggregating to game_team.


        :param game_id_colum:
        :param team_id_column:
        :param features:
        :param weight_column:
        :param target:
        :param model:
        :param multiclassifier:
        :param pred_column:
        """

        self.weight_column = weight_column
        self.game_id_colum = game_id_colum
        self.team_id_column = team_id_column
        self.features = features
        self._target = target
        self.multiclassifier = multiclassifier
        super().__init__(target=self._target, pred_column=pred_column, model=model or LogisticRegression())

    def train(self, df: pd.DataFrame) -> None:

        if len(df[self._target].unique()) > 2:
            logging.warning("target has more than 2 unique values, multiclassifier has therefore been set to True")
            self.multiclassifier = True

        if self._target not in df.columns:
            raise ValueError(f"target {self._target} not in df")
        grouped = self._create_grouped(df)
        self.model.fit(grouped[self.features], grouped[self._target])

    def add_prediction(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Adds prediction to df

        :param df:
        :return: Input df with prediction column
        """

        grouped = self._create_grouped(df)

        if self.multiclassifier:
            grouped[self._pred_column] = self.model.predict_proba(grouped[self.features]).tolist()
        else:
            grouped[self._pred_column] = self.model.predict_proba(grouped[self.features])[:, 1]

        if self.pred_column in df.columns:
            df = df.drop(columns=[self.pred_column])

        df = df.merge(grouped[[self.game_id_colum, self.team_id_column] + [self._pred_column]],
                      on=[self.game_id_colum, self.team_id_column])

        return df

    def _create_grouped(self, df: pd.DataFrame) -> pd.DataFrame:

        if df[self._target].dtype == 'object':
            df.loc[:, self._target] = df[self._target].astype('int')

        if self.weight_column:
            for feature in self.features:
                df = df.assign(**{feature: df[self.weight_column] * df[feature]})

        if self.weight_column:
            grouped = df.groupby([self.game_id_colum, self.team_id_column]).agg({
                **{feature: 'sum' for feature in self.features},
                self._target: 'mean',
                self.weight_column: 'sum',
            }).reset_index()
            for feature in self.features:
                grouped[feature] = grouped[feature] / grouped[self.weight_column]

            grouped.drop(columns=[self.weight_column], inplace=True)

        else:
            grouped = df.groupby([self.game_id_colum, self.team_id_column]).agg({
                **{feature: 'sum' for feature in self.features},
                self._target: 'mean',
            }).reset_index()

        grouped[self._target] = grouped[self._target].astype('int')
        return grouped


class SKLearnClassifierWrapper(BaseMLWrapper):

    def __init__(self,
                 features: list[str],
                 target: Optional[str] = PredictColumnNames.TARGET,
                 model: Optional = None,
                 multiclassifier: bool = False,
                 pred_column: Optional[str] = "prob",
                 column_names: Optional[ColumnNames] = None,
                 categorical_features: Optional[list[str]] = None
                 ):
        self.features = features
        self._target = target
        self.multiclassifier = multiclassifier
        self.column_names = column_names
        self.categorical_features = categorical_features

        super().__init__(target=self._target, pred_column=pred_column, model=model or LogisticRegression())

    def train(self, df: pd.DataFrame) -> None:

        if self.categorical_features is None:
            self.categorical_features = []
            for feature in self.features:
                self.categorical_features.append(feature)
                if df[feature].dtype == "object" or len(df[feature].unique() < 50):
                    logging.warning(
                        f"feature {feature} has been converted to category. To override default behaviour, pass categorical_features into constructor")
                    df.loc[:, feature] = df[feature].astype('category')

        if self.multiclassifier is False and len(df[self._target].unique()) > 2:
            logging.warning("target has more than 2 unique values, multiclassifier has therefore been set to True")
            self.multiclassifier = True

        self.model.fit(df[self.features], df[self._target])

    def add_prediction(self, df: pd.DataFrame) -> pd.DataFrame:

        df = df.copy()

        for categorical_feature in self.categorical_features:
            df.loc[:, categorical_feature] = df[categorical_feature].astype('category')
        if self.multiclassifier:
            df[self._pred_column] = self.model.predict_proba(df[self.features]).tolist()
        else:
            df[self._pred_column] = self.model.predict_proba(df[self.features])[:, 1]
        return df



import copy
import logging
from typing import Optional, Union

import pandas as pd

from player_performance_ratings.predictor.match_predictor import MatchPredictor
from player_performance_ratings.ratings.match_generator import convert_df_to_matches
from player_performance_ratings.scorer import BaseScorer

from player_performance_ratings.tuner.match_predictor_factory import MatchPredictorFactory
from player_performance_ratings.tuner.rating_generator_tuner import RatingGeneratorTuner
from player_performance_ratings.tuner.transformer_tuner import TransformerTuner

logging.basicConfig(level=logging.INFO)


class MatchPredictorTuner():

    """
    Given a search range, it will identify the best combination of pre_transformers, rating_generators and post_transformers.
    Using optuna, it will tune the hyperparameters of each transformer.
    """


    def __init__(self,
                 scorer: BaseScorer,
                 match_predictor_factory: MatchPredictorFactory,
                 pre_transformer_tuner: Optional[TransformerTuner] = None,
                 rating_generator_tuners: Optional[Union[list[RatingGeneratorTuner], RatingGeneratorTuner]] = None,
                 post_transformer_tuner: Optional[TransformerTuner] = None,
                 fit_best: bool = True
                 ):

        """
        :param scorer:
            The class object that scores the predictions
        :param match_predictor_factory:
            The factory that creates the MatchPredictor.
            Contains the parameters to create the MatchPredictor if no parameter-tuning was done.
            Based on hyperparameter-tuning the final way the match-predictor is generated will be based on a combination of the parameters in the factory and the tuned parameters.
        :param pre_transformer_tuner:
            The tuner that tunes the hyperparameters of the pre_transformers.
            If left none or as [], the pre_transformers in the match_predictor_factory will be used.
        :param rating_generator_tuners:
            The tuner that tunes the hyperparameters of the rating_generators.
            If left none or as [], the rating_generators in the match_predictor_factory will be used.
        :param post_transformer_tuner:
            The tuner that tunes the hyperparameters of the post_transformers.
            If left none or as [], the post_transformers in the match_predictor_factory will be used.
        :param fit_best: Whether to refit the match_predictor with the entire training use data using best hyperparameters
        """

        self.scorer = scorer
        self.match_predictor_factory = match_predictor_factory
        self.pre_transformer_tuner = pre_transformer_tuner or []
        self.rating_generator_tuners = rating_generator_tuners or []
        self.post_transformer_tuner = post_transformer_tuner or []
        if isinstance(self.rating_generator_tuners, RatingGeneratorTuner):
            self.rating_generator_tuners = [self.rating_generator_tuners]
        self.fit_best = fit_best

        if not self.pre_transformer_tuner and not self.rating_generator_tuners:
            raise ValueError("If no pre_transformer_tuner is provided, rating_generator_tuners must be provided")

    def tune(self, df: pd.DataFrame) -> MatchPredictor:

        column_names = self.match_predictor_factory.column_names

        best_pre_transformers = copy.deepcopy(self.match_predictor_factory.pre_transformers)
        if self.pre_transformer_tuner:
            logging.info("Tuning PreTransformers")
            best_pre_transformers = self.pre_transformer_tuner.tune(df,
                                                                    match_predictor_factory=self.match_predictor_factory,
                                                                    scorer=self.scorer)
            for pre_rating_transformer in best_pre_transformers:
                df = pre_rating_transformer.transform(df)


        matches = convert_df_to_matches(df=df, column_names=column_names)

        best_rating_generators = copy.deepcopy(self.match_predictor_factory.rating_generators)
        for rating_idx, rating_generator_tuner in enumerate(self.rating_generator_tuners):
            tuned_rating_generator = rating_generator_tuner.tune(df=df, matches=matches, rating_idx=rating_idx,
                                                                 scorer=self.scorer,
                                                                 match_predictor_factory=self.match_predictor_factory)
            best_rating_generators[rating_idx] = tuned_rating_generator

        best_post_transformers = copy.deepcopy(self.match_predictor_factory.post_transformers)
        if self.post_transformer_tuner:
            logging.info("Tuning PostTransformers")
            best_post_transformers = self.post_transformer_tuner.tune(df,
                                                                      match_predictor_factory=self.match_predictor_factory,
                                                                      scorer=self.scorer,
                                                                      pre_rating_transformers=best_pre_transformers,
                                                                      rating_generators=best_rating_generators,
                                                                      )
            for post_rating_transformer in best_post_transformers:
                df = post_rating_transformer.transform(df)

        best_match_predictor = MatchPredictor(column_names=column_names,
                                              rating_generators=best_rating_generators,
                                              pre_rating_transformers=best_pre_transformers,
                                              post_rating_transformers=best_post_transformers,
                                              predictor=self.match_predictor_factory.predictor)
        if self.fit_best:
            best_match_predictor.generate_historical(df=df, matches=matches, store_ratings=True)

        return best_match_predictor



import joblib


from player_performance_ratings.examples.utils import load_nba_game_player_data

from sklearn.metrics import log_loss

from player_performance_ratings import PredictColumnNames
from player_performance_ratings.data_structures import ColumnNames
from player_performance_ratings.predictor.match_predictor import MatchPredictor

from player_performance_ratings.ratings.enums import RatingColumnNames
from player_performance_ratings.ratings.opponent_adjusted_rating.rating_generator import OpponentAdjustedRatingGenerator

column_names = ColumnNames(
    team_id='team_id',
    match_id='game_id',
    start_date="start_date",
    player_id="player_name",
    performance="won",

)
df = load_nba_game_player_data()
df[PredictColumnNames.TARGET] = df['won']
df = df.sort_values(by=[column_names.start_date, column_names.match_id, column_names.team_id, column_names.player_id])

df = (
    df.assign(team_count=df.groupby(column_names.match_id)[column_names.team_id].transform('nunique'))
    .loc[lambda x: x.team_count == 2]
)


rating_generator = OpponentAdjustedRatingGenerator(features_created=[RatingColumnNames.RATING_DIFFERENCE])

match_predictor = MatchPredictor(column_names=column_names, rating_generators=rating_generator)
df = match_predictor.generate_historical(df)

print(log_loss(df[PredictColumnNames.TARGET], df[match_predictor.predictor.pred_column]))

joblib.dump(match_predictor, 'match_predictor.pkl')




import copy
import inspect
import logging
from dataclasses import dataclass
from typing import Optional

import pandas as pd

from player_performance_ratings.data_structures import ColumnNames, Match
from player_performance_ratings.predictor.match_predictor import MatchPredictor
from player_performance_ratings.ratings.enums import RatingColumnNames
from player_performance_ratings.ratings.opponent_adjusted_rating.start_rating_generator import \
    StartRatingGenerator
from player_performance_ratings.scorer.score import BaseScorer, LogLossScorer


@dataclass
class LeagueH2H:
    league: str
    opp_league: str
    weight: float
    mean_rating_change: float


class StartLeagueRatingOptimizer():

    def __init__(self, column_names: ColumnNames,
                 match_predictor: MatchPredictor,
                 max_iterations: int = 20,
                 learning_step: int = 20,
                 scorer: Optional[BaseScorer] = None,
                 weight_div: int = 500,
                 indirect_weight: float = 1.5,
                 verbose: int = 1,
                 ):
        self.match_predictor = match_predictor
        self.max_iterations = max_iterations
        self.learning_step = learning_step
        self.weight_div = weight_div
        self.indirect_weight = indirect_weight
        self.column_names = column_names
        self.verbose = verbose
        self._scores = []
        self._league_ratings_iterations = []
        self.scorer = scorer or LogLossScorer(target=self.match_predictor.predictor.target,
                                              pred_column=self.match_predictor.predictor.pred_column)

    def optimize(self, df: pd.DataFrame, matches: Optional[list[Match]] = None) -> dict[str, float]:
        start_rating_generator = self.match_predictor.rating_generator.team_rating_generator.player_rating_generator.start_rating_generator
        league_ratings = start_rating_generator.league_ratings.copy()
        start_rating_params = list(
            inspect.signature(start_rating_generator.__class__.__init__).parameters.keys())[1:]
        non_league_ratings_params = {attr: getattr(start_rating_generator, attr) for attr in
                                     dir(start_rating_generator) if
                                     attr in start_rating_params and attr != 'league_ratings'}

        for iteration in range(self.max_iterations + 1):

            start_rating_generator = StartRatingGenerator(league_ratings=league_ratings, **non_league_ratings_params)

            if iteration == self.max_iterations:
                min_idx = self._scores.index(min(self._scores))
                logging.info(f"best start rating params {self._league_ratings_iterations[min_idx]}")
                return self._league_ratings_iterations[min_idx]

            match_predictor = copy.deepcopy(self.match_predictor)
            match_predictor.rating_generator.team_rating_generator.player_rating_generator.start_rating_generator = start_rating_generator
            df = match_predictor.generate_historical(df=df, store_ratings=False)
            score = self.scorer.score(df)

            if not league_ratings:
                league_ratings = match_predictor.rating_generator.team_rating_generator.player_rating_generator.start_rating_generator.league_ratings

            self._league_ratings_iterations.append(league_ratings)
            self._scores.append(score)
            if self.verbose:
                logging.info(f"iteration {iteration} finished. Score: {score}. best startings {league_ratings}")

            league_rating_changes = (df
            .groupby([RatingColumnNames.PLAYER_LEAGUE, RatingColumnNames.OPPONENT_LEAGUE])
            .agg(
                {
                    RatingColumnNames.PLAYER_RATING_CHANGE: 'mean',
                    self.column_names.player_id: 'count'
                }
            )
            .reset_index()
            .rename(
                columns={
                    RatingColumnNames.PLAYER_RATING_CHANGE: 'mean_rating_change',
                    self.column_names.player_id: 'count'
                })
            )
            leagues = league_rating_changes[RatingColumnNames.PLAYER_LEAGUE].unique().tolist()

            league_opp_league_h2hs: dict[str, dict[str, LeagueH2H]] = {}
            league_to_played_against_leagues = {}

            for league in leagues:
                league_to_played_against_leagues[league] = \
                    league_rating_changes[league_rating_changes[RatingColumnNames.PLAYER_LEAGUE] == league][
                        RatingColumnNames.OPPONENT_LEAGUE].unique().tolist()

                league_opp_league_h2hs[league] = {}
                for opp_league in leagues:
                    if league == opp_league:
                        continue

                    rows = league_rating_changes[(league_rating_changes[RatingColumnNames.PLAYER_LEAGUE] == league) &
                                                 (league_rating_changes[
                                                      RatingColumnNames.OPPONENT_LEAGUE] == opp_league)]
                    if len(rows) == 0:
                        weight = 0
                        mean_rating_change = 0
                    else:
                        count = rows['count'].iloc[0]
                        weight = min(1, count / self.weight_div)
                        mean_rating_change = rows['mean_rating_change'].iloc[0]

                    league_h2h = LeagueH2H(
                        weight=weight,
                        mean_rating_change=mean_rating_change,
                        league=league,
                        opp_league=opp_league,
                    )
                    league_opp_league_h2hs[league][opp_league] = league_h2h

            final_league_opp_league_h2hs: dict[str, dict[str, LeagueH2H]] = {}

            league_sum_final_weights = {}
            for league, opp_league_h2h in league_opp_league_h2hs.items():

                league_sum_final_weights[league] = 0
                final_league_opp_league_h2hs[league] = {}
                for opp_league, h2h in opp_league_h2h.items():

                    if h2h.weight < 1:
                        league_against = league_to_played_against_leagues[league]
                        opp_played_against = league_to_played_against_leagues[opp_league]
                        shared_leagues = [l for l in league_against if
                                          l in opp_played_against and l not in (league, opp_league)]

                        relative_shared_weighted_mean = 0
                        shared_weight = 0
                        for shared_league in shared_leagues:
                            opp_vs_shared_h2h = league_opp_league_h2hs[opp_league][shared_league]
                            league_vs_shared_h2h = league_opp_league_h2hs[league][shared_league]

                            sum_rating_change = league_vs_shared_h2h.mean_rating_change - opp_vs_shared_h2h.mean_rating_change

                            relative_shared_weighted_mean += min(
                                opp_vs_shared_h2h.weight, league_vs_shared_h2h.weight) * sum_rating_change
                            shared_weight += min(
                                opp_vs_shared_h2h.weight, league_vs_shared_h2h.weight)

                    weighted_mean = h2h.mean_rating_change * h2h.weight + min(1,
                                                                              shared_weight * self.indirect_weight) * relative_shared_weighted_mean
                    final_weight = min(1, h2h.weight + shared_weight)
                    final_league_h2h = LeagueH2H(
                        weight=final_weight,
                        mean_rating_change=weighted_mean,
                        league=league,
                        opp_league=opp_league,
                    )
                    final_league_opp_league_h2hs[league][opp_league] = final_league_h2h
                    league_sum_final_weights[league] += final_weight

            for league, final_opp_league_h2h in final_league_opp_league_h2hs.items():

                if league not in league_ratings:
                    continue
                start_rating_rating_change = 0

                for opp_league, final_h2h in final_opp_league_h2h.items():
                    start_rating_rating_change += final_h2h.weight / league_sum_final_weights[
                        league] * final_h2h.mean_rating_change * self.learning_step

                league_ratings[league] += start_rating_rating_change

        return league_ratings



import logging
from typing import Optional, Union

import pandas as pd

from player_performance_ratings.transformation.base_transformer import BaseTransformer


def create_output_column_by_game_group(data: pd.DataFrame, feature_name: str,
                                       weight_column: str, game_id: str, granularity: list[str]) -> pd.DataFrame:
    if weight_column:
        data = data.assign(**{f'weighted_{feature_name}': data[feature_name] * data[weight_column]})
        data = data.assign(
            **{"sum_weighted": data.groupby(granularity + [game_id])[weight_column].transform("sum")})
        data = data.assign(**{feature_name: data[f'weighted_{feature_name}'] / data['sum_weighted']})
        data = data.groupby(granularity + [game_id])[feature_name].sum().reset_index()

    else:
        data = data.groupby(granularity + [game_id])[feature_name].mean().reset_index()
    return data


class LagTransformation(BaseTransformer):

    def __init__(self,
                 feature_names: list[str],
                 lag_length: int,
                 granularity: Union[list[str], str],
                 game_id: Optional[str] = None,
                 weight_column: Optional[str] = None,
                 df: Optional[pd.DataFrame] = None,
                 prefix: str = 'lag_'
                 ):

        """
        :param
            feature_names: Which features to lag

        :param lag_length:
            Number of lags

        :param granularity:
            Columns to group by before lagging. E.g. player_id or [player_id, position].
             In the latter case it will get the lag for each player_id and position combination.

        :param game_id:
            Column name of game_id.
            If there are more multiple rows per granularity per game_id and you want to add the lag from the prior games, set game_id
            This will calculate the mean of the features per game and the lag will be the prior means

        :param weight_column:
            Only used if game_id is set.
            Will calculate weighted mean of the features per game and the lag will be the prior weighted means.
            This is useful when working with partial game-data of different lengths.
            In that case it can beneficial to set the game-length as weight_colum.

        :param df: Optional parameter to pass in a dataframe to calculate the lag on.
            If not passed in, it will use the dataframe passed in the transform method.
            This is useful if you want to calculate the lag on a different dataframe than the one you want to transform.
            Will merge the two dataframes on game_id and granularity.

        :param prefix:
            Prefix for the new lag columns
        """

        self.feature_names = feature_names
        self.game_id = game_id
        self.granularity = granularity or []
        if isinstance(self.granularity, str):
            self.granularity = [self.granularity]
        self.lag_length = lag_length
        self.weight_column = weight_column
        self.df = df
        self.prefix = prefix
        self._output_feature_names = []

        if self.df is not None and self.game_id is None:
            raise ValueError('If passing in a dataframe to calculate the lag on, you need to set game_id')

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:

        if self.df is not None:
            data = self.df
        else:
            data = df


        for feature_name in self.feature_names:
            for lag in range(1, self.lag_length + 1):
                output_column_name = f'{self.prefix}{lag}_{feature_name}'
                if output_column_name in data.columns:
                    output_column_name += '_1'
                    logging.warning(f'Column {output_column_name} already exists, renaming to {output_column_name}')

                self._output_feature_names.append(output_column_name)

                if self.game_id:
                    data = create_output_column_by_game_group(data=data, feature_name=feature_name,
                                                              weight_column=self.weight_column, game_id=self.game_id,
                                                              granularity=self.granularity)

                data = data.assign(**{output_column_name: data.groupby(self.granularity)[feature_name].shift(lag)})

        if self.game_id is not None:
            df = df.merge(data[ self._output_feature_names + self.granularity + [self.game_id]],
                          on=self.granularity + [self.game_id], how='left')
        else:
            df = data

        return df

    @property
    def features_created(self) -> list[str]:
        return self._output_feature_names

class RollingMeanTransformation(BaseTransformer):

    def __init__(self,
                 feature_names: list[str],
                 window: int,
                 granularity: Union[list[str], str],
                 min_periods: int = 1,
                 game_id: Optional[str] = None,
                 weight_column: Optional[str] = None,
                 df: Optional[pd.DataFrame] = None,
                 prefix: str = 'rolling_mean_'):
        """

        :param feature_names:
            Features to create rolling mean for

        :param granularity:
            Columns to group by before rolling mean. E.g. player_id or [player_id, position].
             In the latter case it will get the rolling mean for each player_id and position combination.

        :param window:
            Window size for rolling mean, if 10 will calculate rolling mean over the prior 10 observations

        :param game_id:
            Column name of game_id.
            If there are more multiple rows per granularity per game_id and you want to add the rolling mean per game_id, set game_id
            This will calculate the mean of the features per game and the rolling mean will be calculated on that.

        :param weight_column:
            Only used if game_id is set.
            Will calculate weighted mean of the features per game and the rolling mean will be the prior weighted means.
            This is useful when working with partial game-data of different lengths.
            In that case it can beneficial to set the game-length as weight_colum.

        :param df: Optional parameter to pass in a dataframe to calculate the rolling mean on.
            If not passed in, it will use the dataframe passed in the transform method.
            This is useful if you want to calculate the rolling mean on a different dataframe than the one you want to transform.
            Will merge the two dataframes on game_id and granularity.

        :param prefix:
            Prefix for the new rolling mean columns
        """

        self.feature_names = feature_names
        self.granularity = granularity
        if isinstance(self.granularity, str):
            self.granularity = [self.granularity]
        self.window = window
        self.min_periods = min_periods
        self.weight_column = weight_column
        self.game_id = game_id
        self.df = df
        self.prefix = prefix
        self._output_feature_names = []

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:

        if self.df is not None:
            data = self.df
        else:
            data = df


        for feature_name in self.feature_names:
            output_column_name = f'{self.prefix}{self.window}_{feature_name}'
            if output_column_name in df.columns:
                output_column_name += '_1'
                logging.warning(f'Column {output_column_name} already exists, renaming to {output_column_name}')

            self._output_feature_names.append(output_column_name)

            if self.game_id:
                data = create_output_column_by_game_group(data=data, feature_name=feature_name,
                                                          weight_column=self.weight_column, game_id=self.game_id,
                                                          granularity=self.granularity)

            data = data.assign(**{output_column_name: data.groupby(self.granularity)[feature_name].apply(
                lambda x: x.shift().rolling(self.window, min_periods=self.min_periods).mean())})

            if self.game_id is not None:
                df = df.merge(data[self._output_feature_names + self.granularity + [self.game_id]],
                              on=self.granularity + [self.game_id], how='left')
            else:
                df = data

        return df

    @property
    def features_created(self) -> list[str]:
        return self._output_feature_names



import math
from abc import ABC, abstractmethod
from typing import Optional

from player_performance_ratings.data_structures import PreMatchPlayerRating, PreMatchTeamRating

MATCH_CONTRIBUTION_TO_SUM_VALUE = 1
MODIFIED_RATING_CHANGE_CONSTANT = 1
CERTAIN_SUM = 'certain_sum'


def sigmoid_subtract_half_and_multiply2(value: float, x: float) -> float:
    return (1 / (1 + math.exp(-value / x)) - 0.5) * 2


class PerformancePredictor(ABC):

    @abstractmethod
    def predict_performance(self,
                            player_rating: PreMatchPlayerRating,
                            opponent_team_rating: PreMatchTeamRating,
                            team_rating: PreMatchTeamRating
                            ) -> float:
        pass


class RatingDifferencePerformancePredictor(PerformancePredictor):

    # TODO: Performance prediction based on team-players sharing time with.
    def __init__(self,
                 rating_diff_coef: float = 0.005757,
                 rating_diff_team_from_entity_coef: float = 0.0,
                 team_rating_diff_coef: float = 0.0,
                 max_predict_value: float = 1,
                 participation_weight_coef: Optional[float] = None,
                 mean_participation_weight: Optional[float] = None,
                 ):
        self.rating_diff_coef = rating_diff_coef
        self.rating_diff_team_from_entity_coef = rating_diff_team_from_entity_coef
        self.team_rating_diff_coef = team_rating_diff_coef
        self.max_predict_value = max_predict_value
        self.mean_participation_weight = mean_participation_weight
        self.participation_weight_coef = participation_weight_coef
        if self.participation_weight_coef and mean_participation_weight is None:
            raise ValueError("mean_duration must be passed if duration_column_name is passed")

    def predict_performance(self,
                            player_rating: PreMatchPlayerRating,
                            opponent_team_rating: PreMatchTeamRating,
                            team_rating: PreMatchTeamRating
                            ) -> float:
        rating_difference = player_rating.rating_value - opponent_team_rating.rating_value

        rating_diff_team_from_entity = team_rating.rating_value - player_rating.rating_value
        team_rating_diff = team_rating.rating_value - opponent_team_rating.rating_value


        value = self.rating_diff_coef * rating_difference + \
                self.rating_diff_team_from_entity_coef * rating_diff_team_from_entity + team_rating_diff * self.team_rating_diff_coef
        if self.participation_weight_coef:
            value += value  * self.participation_weight_coef * (player_rating.match_performance.participation_weight - self.mean_participation_weight)

        prediction = (math.exp(value)) / (1 + math.exp(value))
        if prediction > self.max_predict_value:
            return self.max_predict_value
        elif prediction < (1 - self.max_predict_value):
            return (1 - self.max_predict_value)
        return prediction


class RatingMeanPerformancePredictor(PerformancePredictor):

    def __init__(self,
                 coef: float = 0.005757,
                 max_predict_value: float = 1,
                 last_sample_count: int = 1500
                 ):
        self.coef = coef
        self.max_predict_value = max_predict_value
        self.last_sample_count = last_sample_count
        self.sum_ratings = []
        self.sum_rating = 0
        self.rating_count = 0

    def predict_performance(self,
                            player_rating: PreMatchPlayerRating,
                            opponent_team_rating: PreMatchTeamRating,
                            team_rating: PreMatchTeamRating) -> float:

        self.sum_ratings.append(player_rating.rating_value)
        self.rating_count += 1
        self.sum_rating += player_rating.rating_value
        start_index = max(0, len(self.sum_ratings) - self.last_sample_count)
        self.sum_ratings = self.sum_ratings[start_index:]
        #  average_rating = sum(self.sum_ratings) / len(self.sum_ratings)
        historical_average_rating = self.sum_rating / self.rating_count
        net_mean_rating_over_historical_average = player_rating.rating_value * 0.5 + opponent_team_rating.rating_value * 0.5 - historical_average_rating

        value = self.coef * net_mean_rating_over_historical_average
        prediction = (math.exp(value)) / (1 + math.exp(value))
        if prediction > self.max_predict_value:
            return self.max_predict_value
        elif prediction < (1 - self.max_predict_value):
            return (1 - self.max_predict_value)
        return prediction



import logging
from dataclasses import dataclass, field
from typing import Literal, Optional, Union, Any

from optuna.trial import BaseTrial


@dataclass
class ParameterSearchRange:
    name: str
    type: Literal["uniform", "loguniform", "int", "categorical", "discrete_uniform"]
    low: Optional[Union[float, int]] = None
    high: Optional[Union[float, int]] = None
    choices: Optional[list[Any]] = None
    custom_params: dict[str, Any] = field(default_factory=dict)

def add_params_from_search_range(trial: BaseTrial,parameter_search_range: list[ParameterSearchRange], params: dict) -> dict:
    for config in parameter_search_range:
        if config.type == "uniform":
            params[config.name] = trial.suggest_uniform(config.name, low=config.low, high=config.high)
        elif config.type == "loguniform":
            params[config.name] = trial.suggest_loguniform(config.name, low=config.low, high=config.high)
        elif config.type == "int":
            params[config.name] = trial.suggest_int(config.name, low=config.low, high=config.high)
        elif config.type == "categorical":
            params[config.name] = trial.suggest_categorical(config.name, config.choices)
        else:
            logging.warning(f"Unknown type {config.type} for parameter {config.name}")

    return params



from abc import abstractmethod, ABC
from typing import Optional

import pandas as pd


class BaseMLWrapper(ABC):

    def __init__(self, model, target: str, pred_column: Optional[str] = "prob"):
        self.model = model
        self._target = target
        self._pred_column = pred_column

    @abstractmethod
    def train(self, df: pd.DataFrame) -> None:
        pass

    @abstractmethod
    def add_prediction(self, df: pd.DataFrame) -> pd.DataFrame:
        pass

    @property
    def pred_column(self) -> str:
        return self._pred_column

    @property
    def target(self) -> str:
        return self._target

    @property
    def classes_(self) -> list[str]:
        return self.model.classes_


    def set_target(self, new_target_name: str):
        self._target = new_target_name





from .enums import RatingColumnNames, PredictedRatingMethod, InputColumnNames
from .league_identifier import LeagueIdentifier
from .match_generator import convert_df_to_matches
from .time_weight_ratings import BayesianTimeWeightedRating
from .opponent_adjusted_rating import *


import copy
import inspect
from typing import Optional,  Tuple, Union, Literal

import optuna
import pandas as pd
from optuna.samplers import TPESampler
from optuna.trial import BaseTrial

from player_performance_ratings.ratings.enums import RatingColumnNames
from player_performance_ratings.ratings.opponent_adjusted_rating.rating_generator import RatingGenerator
from player_performance_ratings.scorer.score import BaseScorer
from player_performance_ratings.transformation import ColumnWeight
from player_performance_ratings.transformation.base_transformer import BaseTransformer

from player_performance_ratings.tuner.match_predictor_factory import MatchPredictorFactory

from player_performance_ratings.tuner.utils import add_params_from_search_range, ParameterSearchRange

RC = RatingColumnNames


def insert_params_to_common_transformers(object: object, params, parameter_search_range: list[ParameterSearchRange]):
    if object.__class__.__name__ == "ColumnsWeighter":
        sum_weights = sum([params[p.name] for p in parameter_search_range])
        for p in parameter_search_range:
            params[p.name] = params[p.name] / sum_weights

        column_weights = [ColumnWeight(name=p.name, weight=params[p.name], **p.custom_params) for p in
                          parameter_search_range]

        for p in parameter_search_range:
            del params[p.name]
        params['column_weights'] = column_weights

    return params


def add_hyperparams_to_common_transformers(object: object, params: dict[str, Union[float, None, bool, int, str]],
                                           trial: BaseTrial,
                                           parameter_search_range: list[ParameterSearchRange]) -> dict:
    params = add_params_from_search_range(params=params, trial=trial, parameter_search_range=parameter_search_range)

    return insert_params_to_common_transformers(object=object, params=params,
                                                parameter_search_range=parameter_search_range)


class TransformerTuner:

    def __init__(self,
                 transformer_search_ranges: list[Tuple[BaseTransformer, list[ParameterSearchRange]]],
                 pre_or_post: Literal["pre_rating", "post_rating"],
                 n_trials: int = 30,
                 ):
        self.transformer_search_ranges = transformer_search_ranges
        self.pre_or_post = pre_or_post
        self.n_trials = n_trials
        self._scores = []

    def tune(self, df: pd.DataFrame,
             scorer: BaseScorer,
             match_predictor_factory: MatchPredictorFactory,
             pre_rating_transformers: Optional[list[BaseTransformer]] = None,
             rating_generators: Optional[list[RatingGenerator]] = None,
             ) -> list[BaseTransformer]:

        match_predictor_factory = copy.deepcopy(match_predictor_factory)

        def objective(trial: BaseTrial,
                      df: pd.DataFrame,
                      transformer_search_ranges: list[
                          Tuple[BaseTransformer, list[ParameterSearchRange]]],
                      match_predictor_factory: MatchPredictorFactory,
                      scorer: BaseScorer,
                      ) -> float:
            transformers = []
            for transformer, parameter_search_range in transformer_search_ranges:
                transformer_params = list(
                    inspect.signature(transformer.__class__.__init__).parameters.keys())[1:]
                params = {attr: getattr(transformer, attr) for attr in
                          dir(transformer) if
                          attr in transformer_params}
                params = add_hyperparams_to_common_transformers(object=transformer, params=params, trial=trial,
                                                                parameter_search_range=parameter_search_range)
                class_transformer = type(transformer)
                transformers.append(class_transformer(**params))

            # TODO: Fix it properly so it works with pre and post transformers after tuned
            if self.pre_or_post == "pre_rating":
                match_predictor = match_predictor_factory.create(pre_rating_transformers=transformers)
            else:
                match_predictor = match_predictor_factory.create(pre_rating_transformers=pre_rating_transformers,
                                                                 rating_generators=rating_generators,
                                                                 post_rating_transformers=transformers)

            df_with_prediction = match_predictor.generate_historical(df=df, store_ratings=False)
            return scorer.score(df_with_prediction, classes_=match_predictor.classes_)

        best_transformers = []
        direction = "minimize"
        study_name = "optuna_study"
        optuna_seed = 12
        sampler = TPESampler(seed=optuna_seed)
        study = optuna.create_study(direction=direction, study_name=study_name, sampler=sampler)
        callbacks = []
        study.optimize(
            lambda trial: objective(trial, df=df, transformer_search_ranges=self.transformer_search_ranges,
                                    match_predictor_factory=match_predictor_factory, scorer=scorer),
            n_trials=self.n_trials, callbacks=callbacks)

        best_params = study.best_params
        for transformer, search_range in self.transformer_search_ranges:
            param_values = {p.name: best_params[p.name] for p in search_range if p.name in best_params}
            class_transformer = type(transformer)
            transformer_params = list(
                inspect.signature(transformer.__class__.__init__).parameters.keys())[1:]
            params = {attr: getattr(transformer, attr) for attr in
                      dir(transformer) if
                      attr in transformer_params}

            for k, v in param_values.items():
                params[k] = v

            params = insert_params_to_common_transformers(object=transformer, params=params,
                                                          parameter_search_range=search_range)
            best_transformers.append(class_transformer(**params))

        return best_transformers



import pickle

from sklearn.preprocessing import StandardScaler

from player_performance_ratings.examples.utils import load_lol_data
from player_performance_ratings.data_structures import ColumnNames
from player_performance_ratings import MatchPredictor, OpponentAdjustedRatingGenerator
from player_performance_ratings import SKLearnClassifierWrapper

from player_performance_ratings import TeamRatingGenerator

from player_performance_ratings import SkLearnTransformerWrapper, MinMaxTransformer
from player_performance_ratings.ratings.enums import RatingColumnNames
from player_performance_ratings.transformations.pre_transformers import ColumnsWeighter
from player_performance_ratings.tuner import TransformerTuner
from player_performance_ratings.tuner.utils import ParameterSearchRange

column_names = ColumnNames(
    team_id='teamname',
    match_id='gameid',
    start_date="date",
    player_id="playername",
    performance='performance',
    league='league'
)
df = load_lol_data()
df = df.sort_values(by=['date', 'gameid', 'teamname', "playername"])

df = (
    df.loc[lambda x: x.position != 'team']
    .assign(team_count=df.groupby('gameid')['teamname'].transform('nunique'))
    .loc[lambda x: x.team_count == 2]
)

team_rating_generator = TeamRatingGenerator()
rating_generator = OpponentAdjustedRatingGenerator()
predictor = SKLearnClassifierWrapper(features=[RatingColumnNames.RATING_DIFFERENCE], target='result')

player_search_ranges = [
    ParameterSearchRange(
        name='certain_weight',
        type='uniform',
        low=0.7,
        high=0.95
    ),
    ParameterSearchRange(
        name='certain_days_ago_multiplier',
        type='uniform',
        low=0.02,
        high=.12,
    ),
    ParameterSearchRange(
        name='max_days_ago',
        type='uniform',
        low=40,
        high=150,
    ),
    ParameterSearchRange(
        name='max_certain_sum',
        type='uniform',
        low=20,
        high=70,
    ),
    ParameterSearchRange(
        name='certain_value_denom',
        type='uniform',
        low=15,
        high=50
    ),
    ParameterSearchRange(
        name='reference_certain_sum_value',
        type='uniform',
        low=0.5,
        high=5
    ),
    ParameterSearchRange(
        name='rating_change_multiplier',
        type='uniform',
        low=30,
        high=100
    )
]

features = ["result"]
standard_scaler = SkLearnTransformerWrapper(transformer=StandardScaler(), features=features)

pre_transformers = [
    standard_scaler,
]

duration_performance_search_range = []
column_weigher_search_range = [
    ParameterSearchRange(
        name='damagetochampions',
        type='uniform',
        low=0,
        high=0.45
    ),
    ParameterSearchRange(
        name='deaths',
        type='uniform',
        low=0,
        high=.3,
    ),
    ParameterSearchRange(
        name='kills',
        type='uniform',
        low=0,
        high=0.3
    ),
    ParameterSearchRange(
        name='result',
        type='uniform',
        low=0.25,
        high=0.85
    ),
]

features = ["damagetochampions", "result",
            "kills", "deaths"]
standard_scaler = SkLearnTransformerWrapper(transformer=StandardScaler(), features=features)

pre_transformer_search_ranges = [
    (standard_scaler, []),
    (MinMaxTransformer(features=features), []),
    (
        ColumnsWeighter(weighted_column_name=column_names.performance, column_weights=[]),
        column_weigher_search_range),
]

match_predictor = MatchPredictor(
    column_names=column_names,
    predictor=predictor,
    pre_rating_transformers=pre_transformers,
)

start_rating_search_range = [
    ParameterSearchRange(
        name='team_weight',
        type='uniform',
        low=0.12,
        high=.4,
    ),
    ParameterSearchRange(
        name='league_quantile',
        type='uniform',
        low=0.12,
        high=.4,
    ),
    ParameterSearchRange(
        name='min_count_for_percentiles',
        type='uniform',
        low=20,
        high=100,
    )
]

pre_transformer_tuner = TransformerTuner(match_predictor=match_predictor,
                                            pre_transformer_search_ranges=pre_transformer_search_ranges,
                                            n_trials=15
                                            )

player_rating_tuner = PlayerRatingTuner(match_predictor=match_predictor,
                                        search_ranges=player_search_ranges,
                                        n_trials=20
                                        )

start_rating_tuner = StartRatingTuner(column_names=column_names,
                                      match_predictor=match_predictor,
                                      n_trials=15,
                                      search_ranges=start_rating_search_range)

tuner = MatchPredictorTuner(
    pre_transformer_tuner=pre_transformer_tuner,
    team_rating_tuner=player_rating_tuner,
    start_rating_tuner=start_rating_tuner,
    fit_best=True,
)
best_match_predictor = tuner.tune(df=df)
pickle.dump(best_match_predictor, open("models/lol_match_predictor", 'wb'))



import pandas as pd


def load_lol_data():
    return pd.read_parquet("data/subsample_lol_data")


def load_nba_game_player_data() -> pd.DataFrame:
    game_player = pd.read_pickle("data/game_player.pickle")
    game_team = pd.read_pickle("data/game_team.pickle")
    game = pd.read_pickle("data/game.pickle")
    game = game.rename(columns={'minutes': 'game_minutes'})
    game_player = game_player.merge(game_team[['game_id', 'team_id', 'score', 'score_opponent', 'won']],
                                    on=['game_id', 'team_id'], how='inner')
    game_player.loc[game_player['won'] == True, 'won'] = 1
    game_player.loc[game_player['won'] == False, 'won'] = 0
    return game_player.merge(game, on=['game_id'], how='inner')


def load_nba_game_matchup_data() -> pd.DataFrame:
    game = pd.read_pickle("data/game.pickle")
    lineup_matchup = pd.read_pickle("data/lineup_matchup.pickle")
    # lineup_matchup['unique_combination'] = lineup_matchup['game_id'].astype(str) + '-' + lineup_matchup[
    #     'lineup_id_offense'].astype('str') + '-' + lineup_matchup['lineup_id_defense'].astype('str')

    # Assigning a unique integer to each unique combination
    # lineup_matchup['matchup_game_id'] = pd.factorize(lineup_matchup['unique_combination'])[0]
    # lineup_matchup.drop(columns=['unique_combination'], inplace=True)

    lineup_matchup = lineup_matchup[
        ['game_id', 'team_id_offense', 'lineup_id_offense', 'lineup_id_defense', 'points', 'minutes_lineup_matchup',
         'lineup_offense', 'lineup_defense']]
    temp_df = lineup_matchup.copy()
    temp_df['temp_id'] = temp_df['lineup_id_offense']
    temp_df['temp_lineup'] = temp_df['lineup_offense']
    temp_df['lineup_id_offense'] = temp_df['lineup_id_defense']
    temp_df['lineup_offense'] = temp_df['lineup_defense']
    temp_df['lineup_id_defense'] = temp_df['temp_id']
    temp_df['lineup_defense'] = temp_df['temp_lineup']
    temp_df = temp_df.drop('temp_id', axis=1)
    temp_df = temp_df.rename(
        columns={'points': 'points_opponent', 'minutes_lineup_matchup': 'minutes_lineup_matchup_opponent'})
    temp_df = temp_df[
        ['game_id', 'lineup_id_offense', 'lineup_id_defense', 'points_opponent', 'minutes_lineup_matchup_opponent',
         'lineup_defense', 'lineup_offense']]
    # Merging the original DataFrame with the temporary one
    lineup_matchup = lineup_matchup.merge(temp_df,
                                          on=['game_id', 'lineup_id_offense', 'lineup_id_defense', 'lineup_offense',
                                              'lineup_defense'], how='outer')

    lineup_matchup['points'] = lineup_matchup['points'].fillna(0)
    lineup_matchup['points_opponent'] = lineup_matchup['points_opponent'].fillna(0)
    lineup_matchup['minutes_lineup_matchup'] = lineup_matchup['minutes_lineup_matchup']
    lineup_matchup['minutes_lineup_matchup_opponent'] = lineup_matchup['minutes_lineup_matchup_opponent'].fillna(0)

    lineup_matchup['minutes_lineup_matchup'] = lineup_matchup['minutes_lineup_matchup'] + lineup_matchup[
        'minutes_lineup_matchup_opponent']
    lineup_matchup.drop(columns=['minutes_lineup_matchup_opponent'], inplace=True)

    lineup_match_players = lineup_matchup.explode('lineup_offense')
    lineup_match_players = lineup_match_players.rename(columns={'lineup_offense': 'player_id'})
    lineup_match_players['team_minutes'] = lineup_match_players.groupby(['game_id', 'team_id_offense'])[
                                               'minutes_lineup_matchup'].transform('sum') / 5
    lineup_match_players.rename(columns={'team_id_offense': 'team_id', 'lineup_id_offense': 'lineup_id',
                                         'lineup_id_defense': 'lineup_id_opponent'}, inplace=True)

  #  game_player = pd.read_pickle("data/game_player.pickle")

    game_team = pd.read_pickle("data/game_team.pickle")

    game = game.rename(columns={'minutes': 'game_minutes'})

    lineup_match_players = lineup_match_players.merge(
        game_team[['game_id', 'team_id', 'score', 'score_opponent', 'won']], on=['game_id', 'team_id'])

    lineup_match_players.loc[lineup_match_players['won'] == True, 'won'] = 1
    lineup_match_players.loc[lineup_match_players['won'] == False, 'won'] = 0
    lineup_match_players['matchup_id'] = lineup_match_players['game_id'].astype(str) + '-' + lineup_match_players[
        'lineup_id'].astype('str') + '-' + lineup_match_players['lineup_id_opponent'].astype('str')
    lineup_match_players['participation_weight'] = lineup_match_players['minutes_lineup_matchup'] / lineup_match_players[
        'team_minutes']
    lineup_match_players.drop(columns=['team_minutes'], inplace=True)
    return lineup_match_players.merge(game, on=['game_id'], how='inner')



import logging
from abc import ABC, abstractmethod
from typing import Optional

import numpy as np
import pandas as pd

from player_performance_ratings.ratings import convert_df_to_matches
from player_performance_ratings.ratings.opponent_adjusted_rating.team_rating_generator import TeamRatingGenerator
from player_performance_ratings.ratings.enums import RatingColumnNames

from player_performance_ratings.data_structures import Match, PreMatchRating, PreMatchTeamRating, PlayerRating, \
    TeamRating, ColumnNames, TeamRatingChange


class RatingGenerator(ABC):

    @abstractmethod
    def generate(self, matches: Optional[list[Match]] = None, df: Optional[pd.DataFrame] = None,
                 column_names: Optional[ColumnNames] = None) -> dict[RatingColumnNames, list[float]]:
        pass

    @property
    @abstractmethod
    def player_ratings(self) -> dict[str, PlayerRating]:
        pass

    @property
    @abstractmethod
    def team_ratings(self) -> list[TeamRating]:
        pass

    @property
    @abstractmethod
    def features_created(self) -> list[str]:
        pass


class OpponentAdjustedRatingGenerator(RatingGenerator):
    """
    Generates ratings for players and teams based on the match-performance of the player and the ratings of the players and teams.
    Ratings are updated after a match is finished
    """

    def __init__(self,
                 team_rating_generator: TeamRatingGenerator = TeamRatingGenerator(),
                 features_created: Optional[list[str]] = None,
                 ):

        """

        :param team_rating_generator: The class contains the logic for generating and updating team ratings and contains many parameters that can be tuned.
        :param features_names_created: If called by match_predictor, feature_names_created determines which features will be used for prediction.
        """
        self.team_rating_generator = team_rating_generator
        self._features_created = features_created or [RatingColumnNames.RATING_DIFFERENCE, RatingColumnNames.PLAYER_RATING, RatingColumnNames.TEAM_RATING,
                RatingColumnNames.OPPONENT_RATING]
        self.ratings_df = None

    def generate(self, matches: Optional[list[Match]] = None, df: Optional[pd.DataFrame] = None,
                 column_names: Optional[ColumnNames] = None) -> dict[RatingColumnNames, list[float]]:

        """
        Generate ratings by iterating over each match, calculate predicted performance and update ratings after the match is finished.
        Default settin

        :param matches: list of matches. Each match must contain two teams.
        :param df: The dataframe from which the matches were generated. Only needed if you want to store the ratings in the class object in which case the column names must also be passed.
        :param column_names: The column names of the dataframe. Only needed if you want to store the ratings in the class object in which case the df must also be passed.
        :return: A dictionary containing historical match-rating values.
         These ratings can easily be added as new columns to the original dataframe for later model training or exploration
        """

        if matches is None and df is None or matches is None and column_names is None:
            raise ValueError("If matches is not passed, df and column names must be massed")

        if matches is None:
            matches = convert_df_to_matches(df=df, column_names=column_names)

        if df is not None and column_names:
            logging.info(
                "both df and column names are passed, and match-ratings will therefore be stored in the class object")

        elif column_names and df is None:
            logging.warning(
                "Column names is passed but df not - this match-ratings will not be stored in the class object")

        pre_match_player_rating_values = []
        pre_match_team_rating_values = []
        pre_match_opponent_rating_values = []
        team_opponent_leagues = []
        match_ids = []
        player_rating_changes = []
        player_leagues = []
        player_predicted_performances = []
        performances = []

        team_rating_changes = []
        for match_idx, match in enumerate(matches):
            self._validate_match(match)
            match_team_rating_changes = self._create_match_team_rating_changes(match=match)
            team_rating_changes += match_team_rating_changes

            if match_idx == len(matches) - 1 or matches[match_idx + 1].update_id != match.update_id:
                self._update_ratings(team_rating_changes=team_rating_changes)
                team_rating_changes = []

            for team_idx, team_rating_change in enumerate(match_team_rating_changes):
                opponent_team = match_team_rating_changes[-team_idx + 1]
                for player_idx, player_rating_change in enumerate(team_rating_change.players):
                    pre_match_player_rating_values.append(player_rating_change.pre_match_rating_value)
                    pre_match_team_rating_values.append(team_rating_change.pre_match_rating_value)
                    player_predicted_performances.append(player_rating_change.predicted_performance)
                    pre_match_opponent_rating_values.append(opponent_team.pre_match_rating_value)
                    player_rating_changes.append(player_rating_change.rating_change_value)
                    player_leagues.append(player_rating_change.league)
                    team_opponent_leagues.append(match_team_rating_changes[-team_idx + 1].league)
                    match_ids.append(match.id)
                    performances.append(player_rating_change.performance)

        rating_differences = np.array(pre_match_team_rating_values) - (
            pre_match_opponent_rating_values)
        rating_means = np.array(pre_match_team_rating_values) * 0.5 + 0.5 * np.array(
            pre_match_opponent_rating_values)

        if df is not None and column_names:
            self.ratings_df = df[
                [column_names.team_id, column_names.player_id, column_names.match_id]].assign(
                **{
                    RatingColumnNames.RATING_DIFFERENCE: rating_differences,
                    RatingColumnNames.PLAYER_LEAGUE: player_leagues,
                    RatingColumnNames.OPPONENT_LEAGUE: team_opponent_leagues,
                    RatingColumnNames.PLAYER_RATING: pre_match_player_rating_values,
                    RatingColumnNames.PLAYER_RATING_CHANGE: player_rating_changes,
                    RatingColumnNames.TEAM_RATING: pre_match_team_rating_values,
                    RatingColumnNames.OPPONENT_RATING: pre_match_opponent_rating_values,
                    RatingColumnNames.PERFORMANCE: performances,
                    RatingColumnNames.RATING_MEAN: rating_means,
                    RatingColumnNames.PLAYER_PREDICTED_PERFORMANCE: player_predicted_performances
                })

        return {
            RatingColumnNames.RATING_DIFFERENCE: rating_differences,
            RatingColumnNames.PLAYER_LEAGUE: player_leagues,
            RatingColumnNames.OPPONENT_LEAGUE: team_opponent_leagues,
            RatingColumnNames.PLAYER_RATING: pre_match_player_rating_values,
            RatingColumnNames.PLAYER_RATING_CHANGE: player_rating_changes,
            RatingColumnNames.MATCH_ID: match_ids,
            RatingColumnNames.TEAM_RATING: pre_match_team_rating_values,
            RatingColumnNames.OPPONENT_RATING: pre_match_opponent_rating_values,
            RatingColumnNames.RATING_MEAN: rating_means,
            RatingColumnNames.PLAYER_PREDICTED_PERFORMANCE: player_predicted_performances
        }

    def _create_match_team_rating_changes(self, match: Match) -> list[TeamRatingChange]:

        team_rating_changes = []
        pre_match_rating = PreMatchRating(
            id=match.id,
            teams=self._get_pre_match_team_ratings(match=match),
            day_number=match.day_number
        )

        for team_idx, pre_match_team_rating in enumerate(pre_match_rating.teams):
            team_rating_change = self.team_rating_generator.generate_rating_change(day_number=match.day_number,
                                                                                   pre_match_team_rating=pre_match_team_rating,
                                                                                   pre_match_opponent_team_rating=
                                                                                   pre_match_rating.teams[
                                                                                       -team_idx + 1])
            team_rating_changes.append(team_rating_change)

        return team_rating_changes

    def _update_ratings(self, team_rating_changes: list[TeamRatingChange]):

        for team_rating_change in team_rating_changes:
            self.team_rating_generator.update_rating_by_team_rating_change(team_rating_change=team_rating_change)

    def _get_pre_match_team_ratings(self, match: Match) -> list[PreMatchTeamRating]:
        pre_match_team_ratings = []
        for match_team in match.teams:
            pre_match_team_ratings.append(self.team_rating_generator.generate_pre_match_team_rating(
                match_team=match_team, day_number=match.day_number))

        return pre_match_team_ratings

    def _validate_match(self, match: Match):
        if len(match.teams) < 2:
            print(f"{match.id} only contains {len(match.teams)} teams")
            raise ValueError

    @property
    def player_ratings(self) -> dict[str, PlayerRating]:
        return dict(sorted(self.team_rating_generator.player_ratings.items(),
                           key=lambda item: item[1].rating_value, reverse=True))

    @property
    def team_ratings(self) -> list[TeamRating]:
        team_id_ratings: list[TeamRating] = []
        teams = self.team_rating_generator.teams
        player_ratings = self.player_ratings
        for id, team in teams.items():
            team_player_ratings = [player_ratings[p] for p in team.player_ids]
            team_rating_value = sum([p.rating_value for p in team_player_ratings]) / len(team_player_ratings)
            team_id_ratings.append(TeamRating(id=team.id, name=team.name, players=team_player_ratings,
                                              last_match_day_number=team.last_match_day_number,
                                              rating_value=team_rating_value))

        return list(sorted(team_id_ratings,
                           key=lambda team: team.rating_value, reverse=True))

    @property
    def features_created(self) -> list[str]:
        return self._features_created



import joblib

from player_performance_ratings.examples.utils import load_nba_game_player_data
from player_performance_ratings import MatchPredictor
from player_performance_ratings.consts import PredictColumnNames

match_predictor: MatchPredictor = joblib.load('match_predictor.pkl')
column_names = match_predictor.column_names
df = load_nba_game_player_data()
game_ids = df[match_predictor.column_names.match_id].unique().tolist()
df = df[df[match_predictor.column_names.match_id].isin(game_ids[len(game_ids)- 10:])]
df[PredictColumnNames.TARGET] = df['won']
df = df.sort_values(by=[column_names.start_date, column_names.match_id, column_names.team_id, column_names.player_id])

df = (
    df.assign(team_count=df.groupby(column_names.match_id)[column_names.team_id].transform('nunique'))
    .loc[lambda x: x.team_count == 2]
)


match_predictor.predict(df)


from abc import abstractmethod, ABC

import pandas as pd


class BaseTransformer(ABC):

    @abstractmethod
    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        pass


    @property
    @abstractmethod
    def features_created(self) -> list[str]:
        pass





import logging
from typing import List, Optional, Union

import pandas as pd
import pendulum

from player_performance_ratings.consts import PredictColumnNames
from player_performance_ratings.predictor.estimators.base_estimator import BaseMLWrapper

from player_performance_ratings.predictor.estimators.classifier import SKLearnClassifierWrapper
from player_performance_ratings.data_structures import ColumnNames, Match
from player_performance_ratings.ratings.enums import RatingColumnNames
from player_performance_ratings.ratings.league_identifier import LeagueIdentifier
from player_performance_ratings.ratings.match_generator import convert_df_to_matches
from player_performance_ratings.ratings.opponent_adjusted_rating.rating_generator import RatingGenerator
from player_performance_ratings.transformation.base_transformer import BaseTransformer


class MatchPredictor():

    def __init__(self,
                 column_names: Union[ColumnNames, list[ColumnNames]],
                 rating_generators: Union[RatingGenerator, list[RatingGenerator]],
                 pre_rating_transformers: Optional[List[BaseTransformer]] = None,
                 post_rating_transformers: Optional[List[BaseTransformer]] = None,
                 predictor: [Optional[BaseMLWrapper]] = None,
                 train_split_date: Optional[pendulum.datetime] = None,
                 ):

        """

        :param column_names:
        :param rating_generators: A single or a list of RatingGenerators.
        :param
            pre_rating_transformers: An optional list of transformations that take place rating generation.
            This is generally recommended if a more complex performance-value is used to update ratings.
            Although any type of feature engineering that isn't dependant upon the output of the ratings can be performed here.
        :param post_rating_transformers:
            After rating-generation, additional feature engineering can be performed.
        :param predictor:
        :param train_split_date:
        """
        self.rating_generators = rating_generators if isinstance(rating_generators, list) else [rating_generators]
        self.column_names = column_names if isinstance(column_names, list) else [column_names for _ in
                                                                                 self.rating_generators]
        self.pre_rating_transformers = pre_rating_transformers or []
        self.post_rating_transformers = post_rating_transformers or []

        if predictor is None:
            logging.warning(f"predictor is not set. Will use {RatingColumnNames.RATING_DIFFERENCE} as only feature")

        self.predictor = predictor
        if self.predictor is None:
            features = [c.features_created for c in self.pre_rating_transformers][0] if len(self.pre_rating_transformers) > 0 else [] + \
                       [c.features_created for c in self.post_rating_transformers][0] if len(self.post_rating_transformers) > 0 else [] + \
                       [c.features_created for c in self.rating_generators][0]

            self.predictor = SKLearnClassifierWrapper(
                features=features,
                target=PredictColumnNames.TARGET
            )

        self.predictor.set_target(PredictColumnNames.TARGET)
        self.train_split_date = train_split_date

    def generate_historical(self, df: pd.DataFrame, matches: list[Match] = None,
                            store_ratings: bool = True) -> pd.DataFrame:

        if self.predictor.target not in df.columns:
            raise ValueError(
                f"Target {self.predictor.target} not in df columns. Target always needs to be set equal to {PredictColumnNames.TARGET}")

        for pre_rating_transformer in self.pre_rating_transformers:
            df = pre_rating_transformer.transform(df)

        if self.train_split_date is None:
            self.train_split_date = df.iloc[int(len(df) / 1.3)][self.column_names[0].start_date]

        for rating_idx, rating_generator in enumerate(self.rating_generators):

            rating_column_names = self.column_names[rating_idx]

            if matches is None:
                matches = convert_df_to_matches(column_names=rating_column_names, df=df,
                                                league_identifier=LeagueIdentifier())

            if store_ratings:
                match_ratings = rating_generator.generate(matches, df=df, column_names=rating_column_names)
            else:
                match_ratings = rating_generator.generate(matches)
            for rating_feature, values in match_ratings.items():
                if len(self.rating_generators) > 1:
                    rating_feature_str = rating_feature + str(rating_idx)
                else:
                    rating_feature_str = rating_feature
                df[rating_feature_str] = values

        for post_rating_transformer in self.post_rating_transformers:
            df = post_rating_transformer.transform(df)

        train_df = df[df[self.column_names[0].start_date] <= self.train_split_date]

        self.predictor.train(train_df)
        df = self.predictor.add_prediction(df)
        return df

    def predict(self, df: pd.DataFrame) -> pd.DataFrame:
        for pre_rating_transformer in self.pre_rating_transformers:
            df = pre_rating_transformer.transform(df)

        for rating_idx, rating_generator in enumerate(self.rating_generators):
            rating_column_names = self.column_names[rating_idx]

            matches = convert_df_to_matches(column_names=rating_column_names, df=df,
                                            league_identifier=LeagueIdentifier())

            match_ratings = rating_generator.generate(matches, df=df)
            for rating_feature, values in match_ratings.items():

                if len(self.rating_generators) > 1:
                    rating_feature_str = rating_feature + str(rating_idx)
                else:
                    rating_feature_str = rating_feature
                df[rating_feature_str] = values

        for post_rating_transformer in self.post_rating_transformers:
            df = post_rating_transformer.transform(df)

        df = self.predictor.add_prediction(df)
        return df

    @property
    def classes_(self):
        return self.predictor.model.classes_



from __future__ import annotations

import logging
import warnings

warnings.simplefilter(action='ignore', category=FutureWarning)

from typing import Optional
import pandas as pd

from player_performance_ratings.data_structures import MatchTeam, ColumnNames, MatchPerformance, \
    MatchPlayer, Match
from player_performance_ratings.ratings.league_identifier import LeagueIdentifier

HOUR_NUMBER_COLUMN_NAME = "hour_number"


def convert_df_to_matches(df: pd.DataFrame, column_names: ColumnNames,
                          league_identifier: Optional[LeagueIdentifier] = None) -> list[Match]:
    """
    Converts a dataframe to a list of matches.
    Each dataframe row needs to be a unique combination of match_id and player_id.
    The dataframe needs to contain the following columns:
        * player_id
        * team_id
        * match_id
        * start_date
        * performance


    Dataframe needs to be sorted by date, game_id, team_id in ascending order.

    Optionally a column for participation_weight and league can be passed.
    The participation_weight indicates the percentage (as a ratio) of the time played by the player in the match.
    The league column indicates the league of the match.
    If the league_identifier is passed it will identify the league of the match by the players past matches played.
    If not the  league of the match will be equal to the league of the current match
    """

    df_sorted = df.sort_values(
        by=[column_names.start_date, column_names.match_id,
            column_names.team_id, column_names.player_id])

    if not df.equals(df_sorted):
        raise ValueError("df needs to be sorted by date, game_id, team_id in ascending order")

    col_names = column_names
    df[col_names.start_date] = pd.to_datetime(df[col_names.start_date], format='%Y-%m-%d %H:%M:%S')
    try:
        date_time = df[col_names.start_date].dt.tz_convert('UTC')
    except TypeError:
        date_time = df[col_names.start_date].dt.tz_localize('UTC')
    df[HOUR_NUMBER_COLUMN_NAME] = (date_time - pd.Timestamp("1970-01-01").tz_localize('UTC')) // pd.Timedelta(
        '1h')

    league_in_df = False
    if col_names.league in df.columns.tolist():
        if league_identifier is None:
            logging.warning("League column passed but no league_identifier passed. league will be set to None")
        else:
            league_in_df = True

    prev_match_id = None

    data_dict = df.to_dict('records')

    matches = []

    prev_team_id = None
    prev_row: Optional[None, pd.Series] = None
    match_teams = []
    match_team_players = []
    team_league_counts = {}

    for row in data_dict:
        match_id = row[col_names.match_id]
        team_id = row[col_names.team_id]
        if team_id != prev_team_id and prev_team_id != None or prev_match_id != match_id and prev_match_id != None:
            match_team = _create_match_team(team_league_counts=team_league_counts, team_id=prev_team_id,
                                            match_team_players=match_team_players)
            match_teams.append(match_team)
            match_team_players = []
            team_league_counts = {}

        if match_id != prev_match_id and prev_match_id != None:
            match = _create_match(league_in_df=league_in_df, row=prev_row, match_teams=match_teams,
                                  column_names=column_names)
            matches.append(match)
            match_teams = []

        participation_weight = 1.0
        if col_names.participation_weight:
            participation_weight = row[col_names.participation_weight]

        player_id = row[col_names.team_id]
        if col_names.player_id is not None:
            player_id = row[col_names.player_id]


        if col_names.position is not None:
            position = row[col_names.position]
        else:
            position = None

        if league_in_df:
            match_league = row[column_names.league]
            if team_id not in team_league_counts:
                team_league_counts[team_id] = {}
            player_league = league_identifier.identify(player_id=player_id,
                                                       league_match=match_league)
            if player_league not in team_league_counts[team_id]:
                team_league_counts[team_id][player_league] = 0

            team_league_counts[team_id][player_league] += 1
        else:
            player_league = None

        performance = MatchPerformance(
            performance_value=row[col_names.performance],
            participation_weight=participation_weight,
        )

        match_player = MatchPlayer(
            id=player_id,
            league=player_league,
            performance=performance,
            position=position,
        )
        match_team_players.append(match_player)

        prev_match_id = match_id
        prev_team_id = team_id
        prev_row = row

    match_team = _create_match_team(team_league_counts=team_league_counts, team_id=prev_team_id,
                                    match_team_players=match_team_players)
    match_teams.append(match_team)
    match = _create_match(league_in_df=league_in_df, row=df.iloc[len(df) - 1],
                          match_teams=match_teams, column_names=column_names)
    matches.append(match)

    return matches


def _create_match(league_in_df, row: pd.Series, match_teams: list[MatchTeam], column_names: ColumnNames) -> Match:
    match_id = row[column_names.match_id]
    if league_in_df:
        match_league = row[column_names.league]
    else:
        match_league = None

    return Match(
        id=match_id,
        teams=match_teams,
        day_number=int(row[HOUR_NUMBER_COLUMN_NAME] / 24),
        league=match_league,
        update_id=row[column_names.rating_update_id]
    )


def _create_match_team(team_league_counts: dict, team_id: str,
                       match_team_players: list[MatchPlayer]) -> MatchTeam:
    if team_league_counts:
        team_league = max(team_league_counts[team_id], key=team_league_counts[team_id].get)
    else:
        team_league = None
    return MatchTeam(
        id=team_id,
        players=match_team_players,
        league=team_league
    )



from .post_transformers import LagTransformation, RollingMeanTransformation
from .pre_transformers import SkLearnTransformerWrapper, MinMaxTransformer, DiminishingValueTransformer, \
    SklearnEstimatorImputer, ColumnWeight, ColumnsWeighter

